#!/usr/bin/env python3
import os
import yaml
from pequod_agent.docker_manager import DockerManager
from pequod_agent.models.application import parse_application_spec
from pequod_agent.models.manifest import Manifest
from pequod_agent.cluster.registries.base import Registry


class BootstrapRegistry(Registry):

    def __init__(self, bootstrap_files: list):
        self.data = {}
        docker = DockerManager()

        for bootstrap_file in bootstrap_files:
            try:
                app = parse_application_spec(bootstrap_file)
            except ValueError:
                app = None
            if app:
                manifest = docker.get_manifest_from_image(bootstrap_file)
                self.data[app] = manifest.data
            else:
                with open(bootstrap_file) as fd:
                    data = yaml.safe_load(fd)
                instances = data.get('instances', [])
                if not isinstance(instances, list):
                    raise ValueError('Invalid instances in bootstrap file')
                for inst in instances:
                    if not isinstance(inst['instance_name'], str):
                        raise ValueError('Instance name must be a string')
                    instance_key = (inst['repository_name'], inst['application_name'], inst['application_version'])
                    self.data[instance_key] = {'filesystems': inst.get('filesystems', []),
                                               'services': inst.get('services', []),
                                               'network_rules': inst.get('network_rules', []),
                                               'vaults': inst.get('vaults', [])}

    def get_docker_image_name(self, repository_name: str, application_name: str) -> str:
        if repository_name:
            return '{}/{}'.format(repository_name, application_name)
        return [application_name]

    def get_manifest(self, repository_name: str, application_name: str, application_version: str) -> Manifest:
        data = self.data.get((repository_name, application_name, application_version))
        if not data:
            # retry with empty repository_name (we might have tagged our image with a registry tag)
            data = self.data.get(('', application_name, application_version))
        return Manifest(data) if data else None

