#!/usr/bin/env python3
import pequod_agent
import random
import requests
import socket
from logging import getLogger
from netaddr.ip import IPAddress
from pequod_agent.models.node_state import NodeState
from pequod_agent.networking.ip import format_ip
from pequod_agent.networking.ippool import IPPool
from pequod_agent.networking.iptables import ip6tables
from pequod_agent.utilities.system import execute

class NetworkManager:
    def __init__(self, ip_pool: IPPool):
        self.ip_pool = ip_pool
        self.node_state = NodeState()   # NodeState is a singleton that should be already initialized in agent.py
        self.log = getLogger('pequod.network-manager')

    def find_free_container_ip(self):
        pool = self.ip_pool.subpool(0xaacc)

        container_ips = set()
        for inst in self.node_state.instances:
            container_ips.add(inst.ip_address)

        self.log.debug('Container IPs: %s', container_ips)
        while True:
            ip = random.choice(pool)
            if str(ip) not in container_ips:
                return ip

    def find_free_proxy_ip(self):
        pool = self.ip_pool.subpool(0xff11)

        proxy_ips = set()
        for inst in self.node_state.instances:
            proxy_ips.update(inst.ip_addresses)

        while True:
            ip = random.choice(pool)
            if str(ip) not in proxy_ips:
                return ip


class ServiceChecker:
    def __init__(self):
        self.log = getLogger('pequod.service-checker')

    def check_service(self, address: IPAddress, service: dict):
        http_path = service.get('health_check_http_path')
        if http_path:
            return self.check_http(address, service, http_path)
        else:
            return self.check_tcp(address, service)

    def check_http(self, address: IPAddress, service: dict, http_path: str) -> bool:
        """
        Checks if a http path is available
        """
        local_port = service['local_port']
        service_name = service['service_name']
        service_version = service['service_version']
        self.log.debug('Checking HTTP path %s on port %d providing service %s:%s..',
                       http_path, local_port, service_name, service_version)
        headers = {'User-Agent': 'pequod-agent/{} health-check {}/{}'.format(pequod_agent.__version__,
                                                                             service_name, service_version)}
        try:
            response = requests.get('http://{}:{}{}'.format(format_ip(str(address)), local_port, http_path),
                                    headers=headers, timeout=5)
            if response.status_code != 200:
                self.log.debug('GET %s returns %d', http_path, response.status_code)
                return False
        except requests.RequestException:
            return False
        self.log.info('HTTP path %s on port %d providing service %s:%s returns OK', http_path, local_port, service_name,
                      service_version)
        return True

    def check_tcp(self, address: IPAddress, service: dict) -> bool:
        """
        Checks if a TCP port is available
        """
        local_port = int(service['local_port'])
        service_name = service['service_name']
        service_version = service['service_version']
        self.log.debug('Checking TCP port %d providing service %s:%s..', local_port, service_name, service_version)
        # noinspection PyBroadException
        try:
            s = socket.socket(socket.AF_INET6, socket.SOCK_STREAM)
            s.settimeout(1)
            s.connect((str(address), local_port))
            s.close()
        except:
            return False
        self.log.info('TCP port %d providing service %s:%s is open', local_port, service_name, service_version)
        return True


def setup_container_hosts_file(file_name: str, hostname: str, ip, local_services: list):
    """
    Write /etc/hosts for Docker container
    """

    # TODO separate this in prepare and setup???

    entries = {}

    for service in local_services:
        entries[service.local_hostname] = str(service.local_ip)

    entries['localhost'] = '::1'
    entries[hostname] = str(ip)

    with open(file_name, 'w') as fd:
        for name, ip in sorted(entries.items()):
            fd.write(ip)
            fd.write(' ')
            fd.write(name)
            fd.write('\n')


def setup_container_iptables():
    # drop all IPv6 packets by default
    ip6tables('-P INPUT DROP')
    ip6tables('-P OUTPUT DROP')
    # drop all IPv4 packets by default
    execute('iptables -P INPUT DROP')
    execute('iptables -P OUTPUT DROP')
    # allow ICMPv6 (required for connectivity)
    ip6tables('-A INPUT -p icmpv6 -j ACCEPT')
    ip6tables('-A OUTPUT -p icmpv6 -j ACCEPT')
    # allow all loopback traffic (needed for port checking and traffic between the application and HAProxy)
    ip6tables('-A INPUT -i lo -p tcp -j ACCEPT')
    ip6tables('-A OUTPUT -o lo -p tcp -j ACCEPT')


def setup_container_network_devices(container_net_device: str, ip_address: str, gateway_ip: str, prefixlen: int):
    """
    Create container network devices
    """
    address = '{}/{}'.format(ip_address, prefixlen)
    execute('ip link set dev {cont_dev} name eth0', cont_dev=container_net_device)
    execute('ip link set eth0 up')
    execute('ip addr add {address} dev eth0', address=address)
    if gateway_ip:
        execute('ip -6 route add default via {gw} dev eth0', gw=gateway_ip)

def setup_container_network_rules(network_rules: dict, local_networks: list):
    """
    Process application's network rules (defined in manifest)
    """

    for rule in network_rules:
        if rule['type'] == 'allow':
            # TODO: also consider port (but only if it is defined in the rule)
            ip6tables('-A OUTPUT -p {protocol} -d {network} -j ACCEPT', **rule)
            ip6tables('-A INPUT -p {protocol} -s {network} -m conntrack --ctstate ESTABLISHED,RELATED -j ACCEPT',
                      **rule)
        elif rule['type'] == 'expose':
            ip6tables('-A INPUT -p {protocol} --dport {port} -j ACCEPT', **rule)
            ip6tables('-A OUTPUT -p {protocol} --sport {port} -m conntrack --ctstate ESTABLISHED,RELATED -j ACCEPT',
                      **rule)
        elif rule['type'] == 'allow_internet' and rule['value']:
            # explicitly drop access to local networks
            for network in local_networks:
                subnet_ip = network['ipv6_subnet_ip']
                subnet_prefixlen = network['ipv6_subnet_prefixlen']
                subnet = '{}/{}'.format(subnet_ip, subnet_prefixlen)
                ip6tables('-A OUTPUT -d {network} -j DROP', network=subnet)
                ip6tables('-A INPUT -s {network} -j DROP', network=subnet)
            # allow access to everything else
            ip6tables('-A INPUT -j ACCEPT')
            ip6tables('-A OUTPUT -j ACCEPT')


def setup_container_resolvconf(file_name: str):
    """
    Creates an empty resolv.conf
    """
    with open(file_name, 'w') as fd:
        fd.write('# empty on purpose\n')


def setup_networking(user: str):
    """
    Drop everything from this user (on the host)
    """
    execute('iptables -A OUTPUT -m owner --uid-owner {user} -j DROP', user=user)
    ip6tables('-A OUTPUT -m owner --uid-owner {user} -j DROP', user=user)

