#!/usr/bin/env python3
import logging
import resource
import subprocess

# loop containers
# build instance from container
from subprocess import CalledProcessError
import psutil
import re
from pequod_agent.constants.path import INSTANCE_DATA_PATH
from pequod_agent.constants.units import MB
from pequod_agent.constants.resources import DEFAULT_MEMORY_MB
from nsenter import Namespace
from pequod_agent.errors import ValidationError
from pequod_agent.models.manifest import Manifest
from pequod_agent.sysinfo import cpu_total_mhz, cpu_count

MAX_POSIX_USERNAME_LENGTH = 32

PATTERNS = {
    # hard Docker repository name limit:
    # "Invalid namespace name (..), only [a-z0-9_] are allowed, size between 4 and 30"
    'repository_name': re.compile('^$|^[a-z0-9_]{4,30}$'),
    'application_name': re.compile('^[a-z][a-z0-9-]{,199}$'),
    'application_version': re.compile('^[a-zA-Z0-9._-]{1,200}$'),
    'instance_name': re.compile('^[a-z][a-z0-9-]{,31}$'),
    'zone_name': re.compile('^[a-z][a-z0-9_-]{2,150}$'),
}


class Instance:

    def __init__(self):
        self.name = self.command = self.pid = self.container_id = self.node_name = self.lifecycle_phase = \
            self.ip_address = self.repository_name = self.application_name = self.application_version = \
            self.zone_name = ''
        self.memory_mb = self.instance_started_time = 0
        self.scheduler_name = 'UNKNOWN'
        self.manifest = None
        self.resources = {}

    @staticmethod
    def validate_request(request: dict):
        for key, pattern in PATTERNS.items():
            if key not in request:
                raise ValidationError('Missing key {}'.format(key))
            if not pattern.match(request[key]):
                raise ValidationError('Invalid value for {}: "{}" (allowed: {})'.format(key, request[key],
                                                                                        pattern.pattern))

    @property
    def cgroup(self) -> str:
        try:
            return 'docker/{}'.format(self.container_id)
        except AttributeError:
            return None

    @property
    def container_name(self) -> str:
        return 'app-{name}'.format(**vars(self))

    @staticmethod
    def get_name_from_container_name(container_name):
        try:
            return container_name.split('-', 1)[1]
        except IndexError:
            return container_name

    @property
    def data_path(self):
        return INSTANCE_DATA_PATH / self.name

    @property
    def posix_username(self) -> str:
        """
        Get the posix username to be created to run this instance
        """
        return 'app-{}'.format(self.name)[:MAX_POSIX_USERNAME_LENGTH]

    @property
    def public_dict(self) -> dict:
        """
        Generate dict with "public" information, to send to controller, for example
        """
        values = {
            'instance_name': self.name,
            'node_name': self.node_name,
            'lifecycle_phase': self.lifecycle_phase,
            'metadata': {
                'ipv6': self.ip_address,
                'repository_name': self.repository_name,
                'application_name': self.application_name,
                'application_version': self.application_version,
                'zone_name': self.zone_name,
                'instance_started_time': self.instance_started_time
            },
            'resources': self.resources,
            'scheduler_name': self.scheduler_name
        }
        return values


class StartingInstance(Instance):
    def __init__(self, instance_request: dict, manifest: Manifest):
        self.validate_request(instance_request)
        self.manifest = manifest
        self.node_name = instance_request['node_name']
        self.name = instance_request['instance_name']
        self.repository_name = instance_request['repository_name']
        self.application_name = instance_request['application_name']
        self.application_version = instance_request['application_version']
        self.zone_name = instance_request.get('zone_name', 'default')
        # "command" is only supported in --bootstrap atm
        self.command = instance_request.get('command')
        self.ip_address = None
        self.resources = instance_request['resources']
        self.scheduler_name = instance_request.get('scheduler_name', 'UNKNOWN')
        self.instance_started_time = 0
        self.lifecycle_phase = 'STARTING'
        self.container_id = None

    @property
    def memory_mb(self) -> int:
        """
        Instance memory in MB
        """
        return self.resources.get('memory_mb', DEFAULT_MEMORY_MB)


class RunningInstance(Instance):
    """
    Running Instances
    """
    def __init__(self, node_name, container, detailed):
        self.node_name = node_name
        self.container = container
        self.detailed = detailed
        self.container_id = container['Id']
        self.repository_name = ''
        image = container['Image']
        self.application_name, self.application_version = image.rsplit(':', 1)
        if '/' in self.application_name:
            parts = self.application_name.split('/')
            self.repository_name = parts[-2]
            self.application_name = parts[-1]
        self.lifecycle_phase = 'RUNNING'

    @property
    def user_name(self) -> str:
        # NOTE: this returns the uid because we are starting docker with POSIX uid :-P
        return self.detailed['Config']['User']

    @property
    def cpu_vmhz(self) -> int:
        with open('/sys/fs/cgroup/cpu/docker/{}/cpu.cfs_period_us'.format(self.container_id)) as fd:
            cfs_period_us = int(fd.read())

        system_vmhz = cpu_total_mhz()
        cpus = cpu_count()

        with open('/sys/fs/cgroup/cpu/docker/{}/cpu.cfs_quota_us'.format(self.container_id)) as fd:
            quota = int(fd.read())

        # quota = int(container_vmhz / system_vmhz * cpus * cfs_period_us)
        container_vmhz = int(round((quota * system_vmhz) / (cpus * cfs_period_us)))
        return container_vmhz

    @property
    def ip_addresses(self) -> set:
        netns = self.detailed['State']['Pid']
        ips = set()
        with Namespace(netns, 'net'):
            try:
                out = subprocess.check_output(['ip', '-6', '-o', 'addr', 'show', 'eth0'])
                for line in out.splitlines():
                    words = line.decode('ascii').split(None)
                    if words[5] == 'global':
                        ips.add(words[3].split('/')[0])
            except CalledProcessError:
                pass
        return ips

    @property
    def ip_address(self) -> str:
        ips = self.ip_addresses
        if ips:
            logging.debug('IPs: %s', ips)
            for ip in ips:
                # FIXME: dirty hack to find container IP
                if ':aacc:' in ip:
                    return ip
        return None

    @property
    def zone_name(self) -> str:
        try:
            with (self.data_path / 'ZONE_NAME').open() as fd:
                return fd.read().strip()
        except:
            return 'UNKNOWN'

    @property
    def scheduler_name(self) -> str:
        try:
            with (self.data_path / 'SCHEDULER_NAME').open() as fd:
                return fd.read().strip()
        except:
            return 'UNKNOWN'

    @property
    def memory_mb(self) -> int:
        """
        Instance memory in MB
        """
        with open('/sys/fs/cgroup/memory/docker/{}/memory.limit_in_bytes'.format(self.container_id)) as fd:
            return int(fd.read()) // MB

    @property
    def name(self) -> str:
        container_name = self.detailed['Name']
        return self.get_name_from_container_name(container_name)

    @property
    def instance_started_time(self) -> int:
        pid = self.detailed['State']['Pid']
        proc = psutil.Process(pid)
        return int(proc.create_time())

    @property
    def pid(self) -> str:
        return str(self.detailed['State']['Pid'])

    @property
    def resources(self) -> dict:
        resources = {'cpu_vmhz': self.cpu_vmhz, 'memory_mb': self.memory_mb}

        pid = self.detailed['State']['Pid']

        soft, hard = resource.prlimit(pid, resource.RLIMIT_NPROC)
        resources['processes_count'] = hard
        soft, hard = resource.prlimit(pid, resource.RLIMIT_NOFILE)
        resources['filehandles_count'] = hard

        return resources

