#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import os
import signal
from collections import defaultdict
from pequod_agent.containerization.cgroups import CGgroup
from nsenter import Namespace
from pequod_agent.persistence import save, load_all, remove
from pequod_agent.utilities.system import execute
from pequod_agent.utilities.metaclasses import Singleton

HAPROXY_CONFIG_TEMPLATE = """
global
    daemon
    # enable the following lines to get debug log lines in /var/log/haproxy.log
    #log /dev/log    local0
    #log /dev/log    local1 notice

defaults
    #log global
    mode tcp
    #option tcplog
    timeout connect 5s
    timeout client 10m
    timeout server 10m

{service_rules}

listen stats {proxy_ip}:6
    mode http
    stats enable
    stats uri /
"""

log = logging.getLogger('pequod.haproxy')

class HAProxyManager(metaclass=Singleton):
    # WARNING NOT TESTED. HERE BE DRAGONS (Maybe)

    def __init__(self):
        self.proxies = defaultdict(list)
        self.recover()

    def recover(self):
        stored_instances = load_all('haproxy')
        for instance in stored_instances:
            self.register(instance)

    def register(self, proxy):
        save(proxy, collection='haproxy')
        self.proxies[proxy.container_id].append(proxy)

    def remove(self, proxy):
        log.debug('Removing proxy %s', repr(proxy))
        proxy.kill()
        self.proxies[proxy.container_id].remove(proxy)
        remove(proxy, collection='haproxy')
        del proxy

class HAProxy:

    def __init__(self, app_instance, source_ip, proxy_ip, hostname, services):
        self.last_config_hash = None
        self.app_instance = app_instance
        self.cgroup = CGgroup(app_instance.cgroup)
        self.container_id = app_instance.container_id
        self.source_ip = source_ip
        self.proxy_ip = proxy_ip
        self.services = services
        haproxy_path = app_instance.data_path / 'haproxy'
        self.config_dir = haproxy_path / '{hostname}'.format(hostname=hostname)
        self.config_file = self.config_dir / 'haproxy.conf'
        self.pid_file = self.config_dir / 'pidfile'

    @property
    def pid(self) -> str:
        return self.pid_file.open().read().strip()

    def kill(self):
        try:
            os.kill(int(self.pid), signal.SIGTERM)
        except (ProcessLookupError, FileNotFoundError):
            pass
        try:
            self.config_file.unlink()
            self.pid_file.unlink()
        except FileNotFoundError:
            pass
        try:
            self.config_dir.rmdir()
        except FileNotFoundError:
            pass

    def write_config(self) -> bool:
        """
        Generates config file and writes to file if it is different than before.
        Returns True if changed and False otherwise.
        """
        try:
            self.config_dir.mkdir()
        except FileExistsError:
            pass
        log.debug('Writing HAProxy configuration to %s..', self.config_file)
        service_rules = '\n'.join(s.haproxy_rule.format(source_ip=self.source_ip, proxy_ip=self.proxy_ip)
                                  for s in self.services)
        config = HAPROXY_CONFIG_TEMPLATE.format(source_ip=self.source_ip, service_rules=service_rules,
                                                proxy_ip=self.proxy_ip)
        config_hash = hash(config)
        if config_hash != self.last_config_hash:
            config_file = self.config_file.open('w+')
            config_file.write(config)
            self.last_config_hash = config_hash
            return True
        else:
            return False

    def restart(self):
        if self.write_config():
            cmd = 'haproxy -f {config_file} -p {pid_file}'
            try:
                pid = self.pid
                log.debug('Restarting HAProxy using %s..', self.config_file)
                cmd += ' -sf {pid}'
            except FileNotFoundError:
                log.debug('Starting a new HAProxy using %s..', self.config_file)
                pid = None

            with Namespace(self.app_instance.pid, 'net'):
                try:
                    execute(cmd, pid=pid, **vars(self))
                except:
                    self.last_config_hash = None  # do this to try to make it restart later
                    raise
            self.cgroup.append(self.pid, sub_systems=['cpuacct', 'cpu', 'memory'])
            return True
        else:
            log.debug('HAProxy not restarted because configuration did not change')
            return False

