#!/usr/bin/env python3
import asyncio
import logging
import netifaces
from pequod_agent.networking.iptables import ensure
import requests
import time
from collections import defaultdict
from nsenter import Namespace
from pequod_agent.cluster import Cluster
from pequod_agent.constants.net import VETH_PREFIX, BRIDGE_NAME
from pequod_agent.constants.path import NETNS_PATH
from pequod_agent.constants.units import MB
from pequod_agent.docker_manager import DockerManager
from pequod_agent.drydock.filesystem import setup_container_filesystems, setup_container_rootfs
from pequod_agent.drydock.image_downloader import ImageDownloader
from pequod_agent.drydock.limits import setup_cgroups, setup_container_rlimits
from pequod_agent.drydock.networking import NetworkManager, ServiceChecker, setup_container_iptables, \
    setup_container_hosts_file, setup_container_network_devices, setup_container_network_rules, \
    setup_container_resolvconf, setup_networking
from pequod_agent.errors import StartupError
from pequod_agent.models.cluster_state import ClusterState
from pequod_agent.models.instance import Instance
from pequod_agent.networking.ip import wait_for_dad
from pequod_agent.networking.ippool import IPPool
from pequod_agent.networking.routing.haproxy import HAProxy, HAProxyManager
from pequod_agent.networking.routing.service import Service
from pequod_agent.utils import generate_random_name, pretty_print
from pequod_agent.utilities.system import create_posix_user, execute, get_posix_uid


DEFAULT_HEALTH_CHECK_TIMEOUT_SECONDS = 60


class Starter:
    """
    Perform the heavy lifting for application instance startup.
    """

    def __init__(self, ippool: IPPool, cluster_state: ClusterState):
        self.log = logging.getLogger('pequod.starter')
        self.cluster = Cluster()  # this was already initialized in agent
        self.manager = self.cluster.core_manager
        self.registry = self.cluster.registry
        self.docker = DockerManager()
        self.ippool = ippool
        self.cluster_state = cluster_state
        self.image_downloader = ImageDownloader()
        self.network_manager = NetworkManager(ippool)
        self.service_checker = ServiceChecker()

    @asyncio.coroutine
    def start(self, instance: Instance):
        start_time = time.time()
        yield from asyncio.async(self._start(instance))
        now = time.time()
        self.log.info('%s: Successfully started in %.0fs', instance.name, now - start_time)

    @asyncio.coroutine
    def _start(self, instance: Instance):
        self.setup_instance_data_directory(instance)
        image_names = self.registry.get_docker_image_name(instance.repository_name, instance.application_name)
        tag = instance.application_version
        posix_username = instance.posix_username
        image_name = yield from self.image_downloader.pull_image(instance.application_name, image_names, tag)
        create_posix_user(posix_username)
        setup_networking(posix_username)
        cid = self.create_container(image_name, tag, instance)

        instance.container_id = cid
        binds, gateway_ip, container_net_device, local_services = yield from self.prepare_container(instance)

        try:
            yield from self.start_container(instance, binds, gateway_ip, container_net_device, local_services)
        except:
            data = self.docker.inspect_container(cid)
            state = data['State']

            if not state.get('Running'):
                logs = self.docker.logs(cid)
                logs = logs.decode('utf-8', 'replace').split('\n')
                raise StartupError('Docker container {} died with exit code {} during startup.'.format(cid[:12], state.get('ExitCode')), state.get('ExitCode'),
                                   '=== Last 10 lines of application stdout/stderr ===\n' + '\n'.join(logs[-10:]))
            else:

                # noinspection PyBroadException
                try:
                    self.log.info('Stopping container %s..', cid[:12])
                    # NOTE: we are using a hardcoded timeout value here as the shutdown_timeout_seconds
                    # wait period is only necessary for regular shutdowns
                    self.docker.stop(cid, timeout=5)
                except:
                    pass
                # NOTE: we are not removing the container here as we might
                # still want to look into the Docker logs to debug the failed startup
                raise

    @staticmethod
    def setup_instance_data_directory(instance: Instance):
        haproxy_path = instance.data_path / 'haproxy'
        directories = instance.data_path, haproxy_path
        for path in directories:
            try:
                path.mkdir()
            except FileExistsError:
                pass

        # we need to remember the instance's zone
        with (instance.data_path / 'ZONE_NAME').open('w') as fd:
            fd.write(instance.zone_name)

        # we need to remember the scheduler which created the instance
        with (instance.data_path / 'SCHEDULER_NAME').open('w') as fd:
            fd.write(instance.scheduler_name)

    def get_environment_variables(self, instance: Instance) -> dict:
        # the zone name is needed to allow applications to configure differently for each zone
        env = {'ZONE_NAME': instance.zone_name}

        # we inject all resources as e.g. memory is needed to set the right -Xmx JVM setting
        for key, val in instance.resources.items():
            env['RESOURCE_{}'.format(key.upper())] = val

        return env

    def create_container(self, image_name: str, tag: str, instance: Instance):
        self.log.debug('%s: Creating container with %s:%s', instance.name, image_name, tag)
        try:
            # Remove old container if any
            self.docker.remove_container_by_name(instance.container_name)
            self.log.debug('%s: removed old container %s', instance.name, instance.container_name)
        except ValueError:
            pass
        # docker only seems to accept posix user uid
        posix_uid = get_posix_uid(instance.posix_username)
        try:
            volumes = [fs['mountpoint'] for fs in instance.manifest.filesystems]
        except TypeError:
            volumes = []

        env = self.get_environment_variables(instance)

        res = self.docker.create_container('{}:{}'.format(image_name, tag), command=instance.command,
                                           hostname=instance.posix_username, name=instance.container_name,
                                           mem_limit=instance.memory_mb * MB, user=posix_uid, volumes=volumes,
                                           environment=env)

        self.log.debug('%s: %s', instance.name, res)
        return res['Id']

    def setup_container_vaults(self, instance: Instance) -> dict:
        """
        Fetches and extracts vaults to instance data path.
        """
        binds = {}
        if self.manager:
            vaults = self.manager.get_vaults()
        else:
            return {}
        for vault in instance.manifest.vaults:
            vault_name = vault['vault_name']
            vault_mountpoint = vault['mountpoint']
            vault_type = vault['type']
            zone_name = instance.zone_name
            if vault_type == 'directory':

                self.log.info('%s: Setting up %s vault on %s..', instance.name, vault_name, vault_mountpoint)
                directory = instance.data_path / 'filesystems/vaults' / vault_name
                directory.mkdir(parents=True)
                try:
                    vaults.extract_vault(vault_name, zone_name, directory)
                except requests.HTTPError as e:
                    status_code = e.response.status_code
                    if status_code == 404:
                        self.log.error('%s: Vault %s in %s zone not found', instance.name, vault_name, zone_name)
                    else:
                        self.log.error('%s: %d error when fetching vault %s in %s zone: %s',
                                       instance.name, status_code, vault_name, zone_name, e.response.text)
                    raise e
                binds[directory] = {'bind': vault_mountpoint, 'ro': True}
            else:
                raise NotImplementedError('Vault type {} not implemented yet'.format(vault_type))
        return binds

    @asyncio.coroutine
    def prepare_container(self, instance: Instance) -> (dict, str, str, list):
        short_cid = instance.container_id[:12]

        self.log.info('%s: Preparing container %s..', instance.name, short_cid)
        binds = setup_container_filesystems(instance.name, instance.data_path, instance.manifest.filesystems)
        yield from asyncio.sleep(0.1)
        vault_binds = self.setup_container_vaults(instance)
        binds.update(vault_binds)

        address, gateway_ip, container_net_device = self.prepare_container_networking(instance)
        instance.ip_address = address

        local_services = self.get_local_services(instance)

        return binds, gateway_ip, container_net_device, local_services

    @asyncio.coroutine
    def start_container(self, instance, binds, gateway_ip, container_net_device, local_services):
        resources = instance.resources
        short_cid = instance.container_id[:12]

        self.log.info('%s: Starting container %s with %s..', instance.name, short_cid, pretty_print(resources))
        self.docker.start(instance.container_id, network_mode='none', binds=binds)
        data = self.docker.inspect_container(instance.container_id)

        instance.pid = data['State']['Pid']
        setup_container_rlimits(instance.pid, resources)

        setup_container_rootfs(instance.posix_username, instance.pid)
        setup_container_hosts_file(data['HostsPath'], data['Config']['Hostname'], instance.ip_address, local_services)
        setup_container_resolvconf(data['ResolvConfPath'])
        yield from self.setup_running_container_networking(instance, gateway_ip, container_net_device, local_services)
        # TODO do this before starting container ??
        setup_cgroups(instance.container_id, resources)
        yield from self.wait_for_service_availability(instance)

    @asyncio.coroutine
    def wait_for_service_availability(self, instance: Instance):
        # application manifest specifies which services to actually check and wait for (or which to ignore)
        provided_services = [service for service in instance.manifest.services if service['type'] == 'provide']
        services_to_check = [service for service in provided_services if service.get('health_check', True)]
        ports_to_check = set([service['local_port'] for service in services_to_check])
        # max_wait is configurable in application manifest
        timeouts = [service.get('health_check_timeout_seconds', DEFAULT_HEALTH_CHECK_TIMEOUT_SECONDS)
                    for service
                    in services_to_check]
        # disable max inspection because PyCharm still doesn't recognize default as a valid keyword of max
        # noinspection PyArgumentList
        max_wait = max(timeouts, default=DEFAULT_HEALTH_CHECK_TIMEOUT_SECONDS)
        self.log.info('%s: Waiting up to %ds for availability of %d service(s) on %d TCP ports..',
                      instance.name, max_wait, len(services_to_check), len(ports_to_check))
        start = time.time()
        while ports_to_check:
            with Namespace(instance.pid, 'net'):
                self.log.debug('%s: Checking ports', instance.name)
                for service in services_to_check:

                    if self.service_checker.check_service(instance.ip_address, service):
                        ports_to_check.remove(service['local_port'])
            if ports_to_check:
                remaining = max_wait - (time.time() - start)
                if remaining <= 0:
                    raise ConnectionRefusedError(
                        'Maximum wait time of {} exceeded while waiting for {} open TCP ports'.format(max_wait, len(
                            ports_to_check)))
                self.log.info('%s: Waiting %d more seconds for %d open TCP ports..',
                              instance.name, round(remaining), len(ports_to_check))
                yield from asyncio.sleep(2)

    def prepare_container_networking(self, instance):
        prefixlen = self.ippool.net.prefixlen
        address = self.network_manager.find_free_container_ip()
        host_dev = generate_random_name(VETH_PREFIX, 4)
        container_device = generate_random_name(VETH_PREFIX, 4)
        gw = netifaces.gateways()['default'].get(netifaces.AF_INET6)
        if gw:
            gateway_ip = gw[0]
        else:
            # we are just using our own host's IPv6 address as a default gateway for the container
            # this will make e.g. NAT64 work in our local test VirtualBox (local nat64 Tayga device)
            gateway_ip = self.ippool.own_ip
            self.log.warning('%s: No IPv6 default gateway found, using own host IP %s', instance.name, gateway_ip)
        self.log.info('%s: Configuring %s/%d..', instance.name, address, prefixlen)
        execute('ip link add {host_dev} type veth peer name {cont_dev}', host_dev=host_dev, cont_dev=container_device)
        execute('brctl addif {bridge} {dev}', bridge=BRIDGE_NAME, dev=host_dev)
        execute('ip link set {host_dev} up', host_dev=host_dev)
        # endregion
        return address, gateway_ip, container_device

    @asyncio.coroutine
    def setup_running_container_networking(self, instance, gateway_ip, container_net_device, local_services):
        prefixlen = self.ippool.net.prefixlen
        self.log.debug('%s: Setting up container networking', instance.name)
        pid = instance.pid
        netpath = NETNS_PATH / str(pid)
        netpath.symlink_to('/proc/{}/ns/net'.format(pid))
        # TODO can we do this on a dummy project and then just start container inside that name space?
        execute('ip link set {dev} netns {pid}', dev=container_net_device, pid=pid)
        with Namespace(pid, 'net'):
            self.log.debug('%s: Dropping all IP traffic by default..', instance.name)
            setup_container_iptables()
            self.log.debug('%s: Setting up network devices..', instance.name)
            setup_container_network_devices(container_net_device, instance.ip_address, gateway_ip, prefixlen)

        if local_services:
            yield from self.setup_container_service_routing(instance, instance.ip_address, local_services)

        if instance.manifest.network_rules:
            local_networks = self.manager.get_networks()
            with Namespace(instance.pid, 'net'):
                self.log.debug('%s: Setting up Network Rules', instance.name)
                setup_container_network_rules(instance.manifest.network_rules, local_networks)

            # policy is already "DROP", just add another last rule for convenience to get "port unreachable"
            # execute('ip6tables -A INPUT -j REJECT')

    def get_local_services(self, instance):
        required_services = []
        host_local_ips = dict()
        for service in instance.manifest.services:
            if service['type'] == 'require':
                service_name = service['service_name']
                service_version = service['service_version']
                local_hostname = service['local_hostname']
                local_port = service['local_port']
                endpoints = self.cluster_state.discover_service(self.registry, self.manager, instance.zone_name,
                                                                service_name, service_version)
                s = Service(service_name, service_version, local_hostname, local_port, endpoints)
                if local_hostname not in host_local_ips:
                    host_local_ips[local_hostname] = self.network_manager.find_free_proxy_ip()
                s.local_ip = host_local_ips[local_hostname]
                required_services.append(s)
        return required_services

    @asyncio.coroutine
    def setup_container_service_routing(self, instance: Instance, source_ip, required_services):
        proxy_manager = HAProxyManager()
        if required_services:
            self.log.info('%s: Configuring required services for %s (PID: %s)..', instance.name,
                          instance.application_name, instance.pid)

            services_by_hostname = defaultdict(list)
            for service in required_services:
                self.log.debug('%s: %s:%s is available on local hostname %s', instance.name, service.name, service.version,
                               service.local_hostname)
                services_by_hostname[service.local_hostname].append(service)

            for hostname, services in services_by_hostname.items():
                self.log.info('%s: Configuring host %s..', instance.name, hostname)
                for service in services:
                    if not service.endpoints:
                        self.log.warning('%s: Required service %s:%s has no endpoints', instance.name, service.name, service.version)

                    # TODO: cleanup duplicate code
                    # configuring the firewall rules here is necessary, but we actually do the same as in agent.py
                    with Namespace(instance.pid, 'net'):
                        for endpoint in service.endpoints:
                            # allow HAProxy to connect to remote (required) service
                            ensure('OUTPUT -p tcp --dport {port} -d {ip} -j ACCEPT', port=endpoint.port, ip=endpoint.ip)
                            # allow remote (required) service to answer HAProxy
                            ensure('INPUT -p tcp --sport {port} -s {ip} -j ACCEPT', port=endpoint.port, ip=endpoint.ip)
                            # this rule is necessary for packet fragmentation
                            ensure('INPUT -p tcp -m conntrack --ctstate RELATED,ESTABLISHED -j ACCEPT')
                proxy_ip = services[0].local_ip
                address = '{}/{}'.format(proxy_ip, self.ippool.net.prefixlen)
                self.log.debug('%s: HAProxy IP %s -> %s', instance.name, hostname, address)
                with Namespace(instance.pid, 'net'):
                    self.log.debug('%s: Adding HAProxy IP for %s on NS %s..', instance.name, hostname, instance.pid)
                    execute('ip addr add {address} dev eth0', address=address)
                    wait_for_dad('eth0')
                self.log.debug('%s: Updating service IPs..', instance.name)
                for service in services:
                    self.log.debug('%s: %s -> %s -> %s', instance.name, service.name, hostname, proxy_ip)
                proxy = HAProxy(instance, source_ip, proxy_ip, hostname, services)
                proxy.restart()
                proxy_manager.register(proxy)
